@@ -32,18 +32,6 @@ __license__   = "LGPL"
 # easily cause circular imports.
 #
 
-# The Accessible that has visual focus.
-#
-locusOfFocus = None
-
-# The currently active window.
-#
-activeWindow = None
-
-# The currently active script.
-#
-activeScript = None
-
 # Used to capture keys to redefine key bindings by the user.
 #
 capturingKeys   = False
@@ -66,13 +54,14 @@ bypassNextCommand = False
 #
 searchQuery = None
 
-# Assists with learn mode (what you enter when you press Insert+F1
-# and exit when you press escape.
-#
-learnModeEnabled = False
-
 # Handle to the Orca Preferences Glade GUI object.
 #
 orcaOS = None
 
-listNotificationsModeEnabled = False
+# The AT-SPI device (needed for key grabs).
+#
+device = None
+
+# A list of AT-SPI key grabs for modifiers (ie, insert)
+#
+grabbedModifiers = {}
